#ifndef OSA_LINUX_CPP_RTL_H
#define OSA_LINUX_CPP_RTL_H

#include "../OsaRtl.h"
#include "OsaLinuxC.h" 

/*---------------------------------------------------------------------------*\
  same structures as struct spinlock, list_head and wait_queue_head in Linux
\*---------------------------------------------------------------------------*/

typedef struct{
	volatile unsigned long lock;
} spinlock_otx;

struct list_head_otx{
	struct list_head_otx *next, *prev;
};
typedef list_head_otx list_head_otx; 

struct wait_queue_head_otx {
	spinlock_otx lock;
	struct list_head_otx task_list;
};

typedef struct wait_queue_head_otx wait_queue_head_otx;


/******************************************************************************
******************************************************************************/
class OsaSpinLockImpl {
public:
  OsaSpinLockImpl()
  {
    OsaSpinLockImpl_Ctor(m_pSLock);
    m_nSavedFlags = 0xabcdef12;
    m_bLocked = false;
  }

  ~OsaSpinLockImpl()
  {
    if (m_bLocked) {
      Unlock();
    }
  }

  void Lock()   
  {
    __u32 flags;
    OsaSpinLockImpl_Lock(m_pSLock, &flags);
    
    m_bLocked = true;
    m_nSavedFlags = flags;  
  }
  
  void Unlock()
  {
    if (m_bLocked) {
      OsaSpinLockImpl_Unlock(m_pSLock, m_nSavedFlags);
    } else {
      OSA_EPRINTF(("OsaLinux: Attempt to unlock a non-locked spinlock\n"));
    }
  }

  bool IsLocked()
  {
    return m_bLocked;
  }

private:
  void       *m_pSLock;
  __u32      m_nSavedFlags;
  bool       m_bLocked;
};


/******************************************************************************
******************************************************************************/
class OsaDpcImpl {
public:
  OsaDpcImpl()
  {
    pDpcFct = (void(*)(void *, void *, void *, void *)) NULL;
    pDpcFctContext = (void *) NULL;
  }

  ~OsaDpcImpl()
  {
  }

  virtual void OsaDpcInitialize(void *pFct, void *pContext)
  {
    pDpcFct = (void(*)(void *, void *, void *, void *)) pFct;
    pDpcFctContext = pContext;
  }

  virtual bool OsaDpcInsertQueue(void *pArg1, void *pArg2)
  {
    (*pDpcFct)(NULL, pDpcFctContext, pArg1, pArg2);
    return true;
  }

private:
  void (*pDpcFct) (void *pDpcContext, void *pContext, void *pArg1, void *pArg2);
  void *pDpcFctContext;
};


/******************************************************************************
******************************************************************************/
class OsaEventImpl {
public:
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)    
    OsaEventImpl()
      : m_pWaitQueue(NULL)
      , m_nEventSet(0)
		{}
  #else
	  OsaEventImpl()
      : m_nEventSet(0)
    {
      m_pWaitQHead = &m_WaitQueHead;
      OsaEventImpl_Ctor(m_pWaitQHead);
    }
  #endif

  ~OsaEventImpl()
  {
    // Release possible waiting threads before destroying
    // This is an arbitrary decision on my part /LME990923
    SetEvent();
  }

  bool SetEvent()
  {
    int nEventSetBefore = m_nEventSet; 
    m_nEventSet = 1;

    #if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)    
      osalinux_wake_up(&m_pWaitQueue);
    #else
      OsaEventImpl_wake_up(m_pWaitQHead);
    #endif

    return nEventSetBefore ? true : false;
  }

  void ClearEvent()
  {
    m_nEventSet = 0;
  }

private:
  friend int OsaWaitForSingleEvent(OsaEventImpl *pEvent, OsaUint32 nWaitMaxMs);

  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)    
    struct wait_queue *m_pWaitQueue;
  #else
    wait_queue_head_otx m_WaitQueHead; 
    void *m_pWaitQHead;  
  #endif

  OsaUint32         m_nEventSet;
  OsaSpinLockImpl   m_access;
};


/******************************************************************************
  A Timeout of zero allows the testing of a set of wait conditions and for 
  conditionally performing any side effects if the wait can be immediately
  satisfied
******************************************************************************/
int OsaWaitForSingleEvent(OsaEventImpl *pEvent, OsaUint32 nWaitMaxMs);

#endif
