#ifndef OSA_RTL_H
#define OSA_RTL_H

#include <Api/OsaDef.h>

// Some of these functions are sometimes implemented directly by the kernel
// but for some configurations they are implemented in OSA
#if defined(_WIN32_WCE)
#include "WindowsCe/OsaWindowsCeRtl.h"
#include "Core/OdDebug.h"
#elif OSA_WIN32
#include "Win32/OsaWin32Rtl.h"
#elif OSA_LINUX
#include "Linux/OsaLinuxRtl.h"
#else
#error "Must define either OSA_LINUX or OSA_WIN32"
#endif

// Executions levels. The execution level determines what OSA services
// can be used. A higher level means more restrictions in what OSA functions can
// be called. All levels can be interrupted/preempted by code running on the same level
// or higher. For EXE_LEVEL_CONTROL and EXE_LEVEL_SYSTEM OdOsaCriticalSections can be used
// to protect shared resources. If a resource is used by EXE_LEVEL_DEVICE code it must be
// protected using DeviceLevelSafe callbacks.
typedef enum OsaExeLevel { 
  OSA_EXE_LEVEL_CONTROL = 0, // The least restrictive level, similar to a user thread.
  OSA_EXE_LEVEL_SYSTEM  = 1, // The same priority as the thread scheduler, hence no thread like functionality is available (e.g. "sleep" and "wait" functions).
  OSA_EXE_LEVEL_DEVICE  = 2  // The highest priority, All other activity in the system is blocked, e.g. in the interrupt handler for a device.
} OsaExeLevel;

typedef enum OsaThreadExitCode {
	OSA_THREAD_EXITCODE_INVALID		= 0,	// Invalid exit code
	OSA_THREAD_EXITCODE_EXITED		= 1, // Thread exited
	OSA_THREAD_EXITCODE_STILLACTIVE	= 2  // Thread e
}OsaThreadExitCode;

typedef enum OsaResultE
{
  OSA_RES_SUCCESS             = 0,
  OSA_RES_FAILED              = -1,
  OSA_RES_TIMEOUT             = -2,
  OSA_RES_NOT_IMPL            = -3,
  OSA_RES_NO_MEMORY           = -4,
  OSA_RES_INVALID_PARAMETER   = -5,
  OSA_RES_NOT_SUPPORTED       = -6,
  OSA_RES_INVALID_BUFFER_SIZE = -7,
  OSA_RES_INVALID_REQUEST     = -8,
  OSA_RES_BUSY                = -9
} OsaResultE;


#define OSA_BRD_IOMEM_AREAS_MAX 2    // Max number of board address areas that we can handle (okay to increase, e.g. to 6, which is PCI max)

#define OSA_FIELD_OFFSET(type, field)    ((long)(char *)&(((type *)0)->field))

#define OSA_MAKEPTR(address) ((void *) (address))

#ifdef __cplusplus
extern "C" {
#endif

//struct OsaIrp;

//typedef struct OsaIrpLink {
//  struct OsaIrp *pIrpNext;
//} OsaIrpLink;

                                                                                                                                                                                                                               
typedef struct OsaIrp {
  OsaUint32   nIoCode;
  void        *pInputBuffer;     // Contains the input struct/buffer
  void        *pOutputBuffer;    // Place for the output struct/buffer, NOTE: space most likely shared with input
  OsaUint32   cbInput;           // Size in bytes of the input struct/buffer
  OsaUint32   cbOutputCapacity;  // Max size allowed for the output struct (size of users output buffer)
  OsaUint32   cbOutput;          // Actual size of the output struct/buffer
  OsaResultE  result;            //
  OsaUint32   nOwnerId;          // File object reference (each process gets it's own file object even if the associated device is the same)

//  int coAssocIrps;
//  struct OsaIrp *pMasterIrp;
//  OsaIrpLink  sLink;
} OsaIrp, *OsaIrpP;


//typedef struct OsaIrpQueue {
//  OsaIrpLink sLink;
//} OsaIrpQueue, *OsaIrpQueueP;

OsaUint32   OsaIrpGetIoCode(IN OsaIrpP pIrp);
OsaVoidP    OsaIrpGetInputBuffer(IN OsaIrpP pIrp);
OsaVoidP    OsaIrpGetOutputBuffer(IN OsaIrpP pIrp);
OsaUint32   OsaIrpGetOwnerId(IN OsaIrpP pIrp);
OsaUint32   OsaIrpGetInputBufferSize(IN OsaIrpP pIrp);
OsaUint32   OsaIrpGetOutputBufferCapacity(IN OsaIrpP pIrp);
//void        OsaIrpMarkIrpPending(IN OsaIrpP pIrp);
OsaResultE  OsaIrpComplete(IN OsaIrpP pIrp, IN OsaResultE eResult, IN OsaUint32 cbOutput);

//OsaIrpP OsaIrpAllocate();
//OsaIrpP OsaIrpAllocateAssocIrp(OsaIrpP pMasterIrp);
//void    OsaIrpDelete(OsaIrpP pIrp);

//void    OsaIrpQueueAdd(IN OsaIrpQueueP pIrpQueue, IN OsaIrpP pIrp);
//OsaIrpP OsaIrpQueueExtract(IN OsaIrpQueueP pIrpQueue);

OsaUint8    OsaInterlockedModifyRegBits8(OsaUint32 pRegister, OsaUint8 nBitsToModify, OsaUint8 nValueOfBits);
OsaUint8    OsaInterlockedModifyPortBits8(OsaUint32 pRegister, OsaUint8 nBitsToModify, OsaUint8 nValueOfBits);

void        OsaInterlockedIncrement(IN OUT OsaInterlockedIntP pValue);
OsaInt32    OsaInterlockedDecrement(IN OUT OsaInterlockedIntP pValue);

void        OsaDelayThreadMicroSecs(OsaInt64 nMicroSecs);
void        OsaStallExecutionMicroSecs(IN OsaUint32 nMicroSecs);

OsaExeLevel OsaGetCurrentExeLevel();

void        OsaQuerySystemTime(OsaDateTime *pDateTime);

OsaUint32   OsaGetSystemTimeTickFrequency(void);

void        OsaSetSystemTimeSource(OsaBool32 bUse1601TypeTimeStamp);

OsaUint32   OsaQueryClockMs(void);

OsaResultE  OsaGetKernelRefForEventHandle(OsaVoidP hEvent, OsaUserEventP *ppUserEvent);

OsaBool32   OsaSetUserEvent(OsaUserEventP pUserEvent);

void        OsaClearUserEvent(OsaUserEventP pUserEvent);

OsaUint64		OsaInt64div32(OsaUint64 divident, OsaUint32 divider);

#if defined (_WIN32_WCE)

char*		_strrev(char *pszUser);

#endif //_WIN32_WCE

#if defined(_TURNSTONE_) 
// Turnstone specific changes

OsaResultE OsaCreateThread(OsaVoidP hThreadHandle,        /* Pointer to Thread handle */
						   OsaVoidP pStartRoutine,        /* Thread Start routine */
						   OsaVoidP pStartContext,        /* Start context for thread*/
						   OsaUint32 dwCreationFlags,  /* 0 = active , 1 = suspended
													      * works only for WINCE   
													      */
						   OsaUint32 dwStackSize,      /* works only for WINCE  */
						   OsaVoidP pThreadId);  /* works only for WINCE  */

// Create a pointer array for allocating 
// Mutex when ever required, upto 10 mutex at a time
// can be supported in the driver


OsaUint32 OsaCreateMutex(OsaUint32  uLevel);                     /* ULevel not used in WINCE can be 
                                                                  * zero in that case. In case
																  * of WIN32 Return value is																  * pMutexObj is 
																  * a handle to mutex object
																  */

OsaResultE OsaReleaseMutex(OsaUint32 uHandle,OsaBool32 uWait);      /* Wait is not counted in CE */
																  


OsaResultE OsaWaitForMutexObject(OsaUint32 uHandle,OsaUint32 uWaitTime);   /* uHandle will be casted to PKMUTEX internally */

						   

OsaResultE OsaGetExitCodeThread(OsaUint32 uHandle,OsaThreadExitCode* uExitCode);

#endif //_TURNSTONE_

#ifdef __linux__
/* linux v2.4 compilation error, but not for strcat() */
/* same segmentation error on /usr/include/asm/string.h even linux/string.h is under "#include" in this file */
	int         OsaStrLen(const char *s);
	void				OsaStrCpy(char *s, char	*t);
	int					OsaStrCmp(const char *s, const char *t);
#else
  #define OsaStrLen strlen
  #define OsaStrCmp strcmp
  #define OsaStrCpy strcpy
#endif

// generic user memory descriptor block structure
//    we use this structure when locking down/accessing
//    user mode memory (non-kernel memory)
typedef struct {
  OsaVoidP    hMemory;            // generic memory descriptor
  OsaVoidP    pUserBuffer;        // address of unlocked (paged) user buffer
  OsaUint32   iUserBufferSize;    // size of user buffer
  OsaVoidP    pLockedBuffer;      // address of locked (non-paged) memory
} OsaUserMemoryDesc;

OsaVoidP OsaLockUserMemory(IN OUT OsaUserMemoryDesc *pOsaUMD, IN OsaVoidP pUserBuffer, IN OsaUint32 uiUserbufferSize);
OsaResultE OsaUnlockUserMemory(IN OUT OsaUserMemoryDesc *pOsaUMD);

// platform type - enabling us to make run-time decisions
typedef enum OsaPlatformE {
  OSA_PLATFORM_UNDEFINED      = 0,
  OSA_PLATFORM_CE_STRONGARM   = 1,
  OSA_PLATFORM_CE_XSCALE      = 2
} OsaPlatformE;

extern OsaPlatformE gOsaPlatform;

#ifdef __cplusplus
}
#endif
#endif









